# ToDo (tribunsky.kir): looks like that idea of dynamic import via imp depending on librdkafka
#                       wasn't the worst idea, cause `if`s causes a lot of static checks problems.
# mypy: disable-error-code="no-redef"

from wunderkafka import librdkafka

from enum import Enum


class BrokerAddressFamily(str, Enum):
    any = 'any'
    v4 = 'v4'
    v6 = 'v6'


class SecurityProtocol(str, Enum):
    plaintext = 'plaintext'
    ssl = 'ssl'
    sasl_plaintext = 'sasl_plaintext'
    sasl_ssl = 'sasl_ssl'


class SslEndpointIdentificationAlgorithm(str, Enum):
    none = 'none'
    https = 'https'


class IsolationLevel(str, Enum):
    read_uncommitted = 'read_uncommitted'
    read_committed = 'read_committed'


class AutoOffsetReset(str, Enum):
    smallest = 'smallest'
    earliest = 'earliest'
    beginning = 'beginning'
    largest = 'largest'
    latest = 'latest'
    end = 'end'
    error = 'error'


class CompressionCodec(str, Enum):
    none = 'none'
    gzip = 'gzip'
    snappy = 'snappy'
    lz4 = 'lz4'
    zstd = 'zstd'


class CompressionType(str, Enum):
    none = 'none'
    gzip = 'gzip'
    snappy = 'snappy'
    lz4 = 'lz4'
    zstd = 'zstd'


class QueuingStrategy(str, Enum):
    fifo = 'fifo'
    lifo = 'lifo'

if librdkafka.__version__ >= (1, 5, 0):
    from enum import Enum
    
    
    class BrokerAddressFamily(str, Enum):
        any = 'any'
        v4 = 'v4'
        v6 = 'v6'
    
    
    class SecurityProtocol(str, Enum):
        plaintext = 'plaintext'
        ssl = 'ssl'
        sasl_plaintext = 'sasl_plaintext'
        sasl_ssl = 'sasl_ssl'
    
    
    class SslEndpointIdentificationAlgorithm(str, Enum):
        none = 'none'
        https = 'https'
    
    
    class IsolationLevel(str, Enum):
        read_uncommitted = 'read_uncommitted'
        read_committed = 'read_committed'
    
    
    class AutoOffsetReset(str, Enum):
        smallest = 'smallest'
        earliest = 'earliest'
        beginning = 'beginning'
        largest = 'largest'
        latest = 'latest'
        end = 'end'
        error = 'error'
    
    
    class CompressionCodec(str, Enum):
        none = 'none'
        gzip = 'gzip'
        snappy = 'snappy'
        lz4 = 'lz4'
        zstd = 'zstd'
    
    
    class CompressionType(str, Enum):
        none = 'none'
        gzip = 'gzip'
        snappy = 'snappy'
        lz4 = 'lz4'
        zstd = 'zstd'
    
    
    class QueuingStrategy(str, Enum):
        fifo = 'fifo'
        lifo = 'lifo'

if librdkafka.__version__ >= (1, 6, 0):
    from enum import Enum
    
    
    class BrokerAddressFamily(str, Enum):
        any = 'any'
        v4 = 'v4'
        v6 = 'v6'
    
    
    class SecurityProtocol(str, Enum):
        plaintext = 'plaintext'
        ssl = 'ssl'
        sasl_plaintext = 'sasl_plaintext'
        sasl_ssl = 'sasl_ssl'
    
    
    class SslEndpointIdentificationAlgorithm(str, Enum):
        none = 'none'
        https = 'https'
    
    
    class IsolationLevel(str, Enum):
        read_uncommitted = 'read_uncommitted'
        read_committed = 'read_committed'
    
    
    class AutoOffsetReset(str, Enum):
        smallest = 'smallest'
        earliest = 'earliest'
        beginning = 'beginning'
        largest = 'largest'
        latest = 'latest'
        end = 'end'
        error = 'error'
    
    
    class CompressionCodec(str, Enum):
        none = 'none'
        gzip = 'gzip'
        snappy = 'snappy'
        lz4 = 'lz4'
        zstd = 'zstd'
    
    
    class CompressionType(str, Enum):
        none = 'none'
        gzip = 'gzip'
        snappy = 'snappy'
        lz4 = 'lz4'
        zstd = 'zstd'
    
    
    class QueuingStrategy(str, Enum):
        fifo = 'fifo'
        lifo = 'lifo'

if librdkafka.__version__ >= (1, 7, 0):
    from enum import Enum
    
    
    class BrokerAddressFamily(str, Enum):
        any = 'any'
        v4 = 'v4'
        v6 = 'v6'
    
    
    class SecurityProtocol(str, Enum):
        plaintext = 'plaintext'
        ssl = 'ssl'
        sasl_plaintext = 'sasl_plaintext'
        sasl_ssl = 'sasl_ssl'
    
    
    class SslEndpointIdentificationAlgorithm(str, Enum):
        none = 'none'
        https = 'https'
    
    
    class IsolationLevel(str, Enum):
        read_uncommitted = 'read_uncommitted'
        read_committed = 'read_committed'
    
    
    class AutoOffsetReset(str, Enum):
        smallest = 'smallest'
        earliest = 'earliest'
        beginning = 'beginning'
        largest = 'largest'
        latest = 'latest'
        end = 'end'
        error = 'error'
    
    
    class CompressionCodec(str, Enum):
        none = 'none'
        gzip = 'gzip'
        snappy = 'snappy'
        lz4 = 'lz4'
        zstd = 'zstd'
    
    
    class CompressionType(str, Enum):
        none = 'none'
        gzip = 'gzip'
        snappy = 'snappy'
        lz4 = 'lz4'
        zstd = 'zstd'
    
    
    class QueuingStrategy(str, Enum):
        fifo = 'fifo'
        lifo = 'lifo'

if librdkafka.__version__ >= (1, 8, 0):
    from enum import Enum
    
    
    class BrokerAddressFamily(str, Enum):
        any = 'any'
        v4 = 'v4'
        v6 = 'v6'
    
    
    class SecurityProtocol(str, Enum):
        plaintext = 'plaintext'
        ssl = 'ssl'
        sasl_plaintext = 'sasl_plaintext'
        sasl_ssl = 'sasl_ssl'
    
    
    class SslEndpointIdentificationAlgorithm(str, Enum):
        none = 'none'
        https = 'https'
    
    
    class IsolationLevel(str, Enum):
        read_uncommitted = 'read_uncommitted'
        read_committed = 'read_committed'
    
    
    class AutoOffsetReset(str, Enum):
        smallest = 'smallest'
        earliest = 'earliest'
        beginning = 'beginning'
        largest = 'largest'
        latest = 'latest'
        end = 'end'
        error = 'error'
    
    
    class CompressionCodec(str, Enum):
        none = 'none'
        gzip = 'gzip'
        snappy = 'snappy'
        lz4 = 'lz4'
        zstd = 'zstd'
    
    
    class CompressionType(str, Enum):
        none = 'none'
        gzip = 'gzip'
        snappy = 'snappy'
        lz4 = 'lz4'
        zstd = 'zstd'
    
    
    class QueuingStrategy(str, Enum):
        fifo = 'fifo'
        lifo = 'lifo'

if librdkafka.__version__ >= (1, 9, 0):
    from enum import Enum
    
    
    class BrokerAddressFamily(str, Enum):
        any = 'any'
        v4 = 'v4'
        v6 = 'v6'
    
    
    class SecurityProtocol(str, Enum):
        plaintext = 'plaintext'
        ssl = 'ssl'
        sasl_plaintext = 'sasl_plaintext'
        sasl_ssl = 'sasl_ssl'
    
    
    class SslEndpointIdentificationAlgorithm(str, Enum):
        none = 'none'
        https = 'https'
    
    
    class SaslOauthbearerMethod(str, Enum):
        default = 'default'
        oidc = 'oidc'
    
    
    class IsolationLevel(str, Enum):
        read_uncommitted = 'read_uncommitted'
        read_committed = 'read_committed'
    
    
    class AutoOffsetReset(str, Enum):
        smallest = 'smallest'
        earliest = 'earliest'
        beginning = 'beginning'
        largest = 'largest'
        latest = 'latest'
        end = 'end'
        error = 'error'
    
    
    class CompressionCodec(str, Enum):
        none = 'none'
        gzip = 'gzip'
        snappy = 'snappy'
        lz4 = 'lz4'
        zstd = 'zstd'
    
    
    class CompressionType(str, Enum):
        none = 'none'
        gzip = 'gzip'
        snappy = 'snappy'
        lz4 = 'lz4'
        zstd = 'zstd'
    
    
    class QueuingStrategy(str, Enum):
        fifo = 'fifo'
        lifo = 'lifo'

if librdkafka.__version__ >= (2, 0, 0):
    from enum import Enum
    
    
    class BrokerAddressFamily(str, Enum):
        any = 'any'
        v4 = 'v4'
        v6 = 'v6'
    
    
    class SecurityProtocol(str, Enum):
        plaintext = 'plaintext'
        ssl = 'ssl'
        sasl_plaintext = 'sasl_plaintext'
        sasl_ssl = 'sasl_ssl'
    
    
    class SslEndpointIdentificationAlgorithm(str, Enum):
        none = 'none'
        https = 'https'
    
    
    class SaslOauthbearerMethod(str, Enum):
        default = 'default'
        oidc = 'oidc'
    
    
    class IsolationLevel(str, Enum):
        read_uncommitted = 'read_uncommitted'
        read_committed = 'read_committed'
    
    
    class AutoOffsetReset(str, Enum):
        smallest = 'smallest'
        earliest = 'earliest'
        beginning = 'beginning'
        largest = 'largest'
        latest = 'latest'
        end = 'end'
        error = 'error'
    
    
    class CompressionCodec(str, Enum):
        none = 'none'
        gzip = 'gzip'
        snappy = 'snappy'
        lz4 = 'lz4'
        zstd = 'zstd'
    
    
    class CompressionType(str, Enum):
        none = 'none'
        gzip = 'gzip'
        snappy = 'snappy'
        lz4 = 'lz4'
        zstd = 'zstd'
    
    
    class QueuingStrategy(str, Enum):
        fifo = 'fifo'
        lifo = 'lifo'

if librdkafka.__version__ >= (2, 1, 0):
    from enum import Enum
    
    
    class BrokerAddressFamily(str, Enum):
        any = 'any'
        v4 = 'v4'
        v6 = 'v6'
    
    
    class SecurityProtocol(str, Enum):
        plaintext = 'plaintext'
        ssl = 'ssl'
        sasl_plaintext = 'sasl_plaintext'
        sasl_ssl = 'sasl_ssl'
    
    
    class SslEndpointIdentificationAlgorithm(str, Enum):
        none = 'none'
        https = 'https'
    
    
    class SaslOauthbearerMethod(str, Enum):
        default = 'default'
        oidc = 'oidc'
    
    
    class IsolationLevel(str, Enum):
        read_uncommitted = 'read_uncommitted'
        read_committed = 'read_committed'
    
    
    class AutoOffsetReset(str, Enum):
        smallest = 'smallest'
        earliest = 'earliest'
        beginning = 'beginning'
        largest = 'largest'
        latest = 'latest'
        end = 'end'
        error = 'error'
    
    
    class CompressionCodec(str, Enum):
        none = 'none'
        gzip = 'gzip'
        snappy = 'snappy'
        lz4 = 'lz4'
        zstd = 'zstd'
    
    
    class CompressionType(str, Enum):
        none = 'none'
        gzip = 'gzip'
        snappy = 'snappy'
        lz4 = 'lz4'
        zstd = 'zstd'
    
    
    class QueuingStrategy(str, Enum):
        fifo = 'fifo'
        lifo = 'lifo'

if librdkafka.__version__ >= (2, 2, 0):
    from enum import Enum
    
    
    class BrokerAddressFamily(str, Enum):
        any = 'any'
        v4 = 'v4'
        v6 = 'v6'
    
    
    class SecurityProtocol(str, Enum):
        plaintext = 'plaintext'
        ssl = 'ssl'
        sasl_plaintext = 'sasl_plaintext'
        sasl_ssl = 'sasl_ssl'
    
    
    class SslEndpointIdentificationAlgorithm(str, Enum):
        none = 'none'
        https = 'https'
    
    
    class SaslOauthbearerMethod(str, Enum):
        default = 'default'
        oidc = 'oidc'
    
    
    class ClientDnsLookup(str, Enum):
        use_all_dns_ips = 'use_all_dns_ips'
        resolve_canonical_bootstrap_servers_only = 'resolve_canonical_bootstrap_servers_only'
    
    
    class IsolationLevel(str, Enum):
        read_uncommitted = 'read_uncommitted'
        read_committed = 'read_committed'
    
    
    class AutoOffsetReset(str, Enum):
        smallest = 'smallest'
        earliest = 'earliest'
        beginning = 'beginning'
        largest = 'largest'
        latest = 'latest'
        end = 'end'
        error = 'error'
    
    
    class CompressionCodec(str, Enum):
        none = 'none'
        gzip = 'gzip'
        snappy = 'snappy'
        lz4 = 'lz4'
        zstd = 'zstd'
    
    
    class CompressionType(str, Enum):
        none = 'none'
        gzip = 'gzip'
        snappy = 'snappy'
        lz4 = 'lz4'
        zstd = 'zstd'
    
    
    class QueuingStrategy(str, Enum):
        fifo = 'fifo'
        lifo = 'lifo'




"""This module contains common datastructures which is used in the package."""

from typing import Optional
from dataclasses import dataclass

from wunderkafka.time import ts2dt


@dataclass(frozen=True)
class Timestamp(object):
    value: int

    def __str__(self) -> str:
        return '{0}: {1:.2f} ({2})'.format(self.__class__.__name__, self.value, ts2dt(self.value))


@dataclass(frozen=True)
class Offset(object):
    value: int


# ToDo (ka.tribunskii): compose header & meta in symmetrical way
@dataclass(frozen=True)
class ParsedHeader(object):
    protocol_id: int
    meta_id: Optional[int]
    schema_id: Optional[int]
    schema_version: Optional[int]
    size: int


@dataclass(frozen=True)
class SchemaMeta(object):
    topic: str
    is_key: bool
    header: ParsedHeader

    @property
    def subject(self) -> str:
        """
        Return topic mapped to schema registry entity.

        :return:            String which should be used as path in schema registry's url corresponding to a given topic.
        """
        # Confluent
        if self.header.protocol_id == 0:
            suffix = '_key' if self.is_key else '_value'
        # Cloudera
        else:
            suffix = ':k' if self.is_key else ''
        return '{0}{1}'.format(self.topic, suffix)


# ToDo: (tribunsky.kir): add cross-validation of inveriants on the model itself?
@dataclass(frozen=True)
class SRMeta(object):
    """Meta which is retrieved after schema registration."""
    # Confluent always has schemd_id, but one of cloudera protocols doesn't use it
    schema_id: Optional[int]
    # Confluent has schema_version, but serdes works around schema_id which is unique identifier there.
    schema_version: Optional[int]
    # Confluent doesn't have metaId
    meta_id: Optional[int] = None


@dataclass(frozen=True)
class SchemaDescription(object):
    """
    Class to allow contract extension between moving parts of (de)serialization.

    Usually schema is represented by text, but separate class adds abstraction.
    """

    text: str



from pathlib import Path

import pytest
from pydantic import Field, BaseModel

from wunderkafka.time import now
from wunderkafka.tests import TestProducer, TestHTTPClient
from wunderkafka.serdes.avro import AvroModelSerializer, ConfluentClouderaHeadersHandler
from wunderkafka.serdes.store import AvroModelRepo
from wunderkafka.schema_registry import SimpleCache, ClouderaSRClient
from wunderkafka.producers.constructor import HighLevelSerializingProducer


class Signal(BaseModel):
    ts: int = Field(default_factory=now)
    source: str = 'test'
    type: str = 'string'
    id: str = 'string'
    value: str = 'NA'

    class Meta:
        name = 'SignalsTest'


class DefaultKey(BaseModel):
    ts: int = Field(default_factory=now)

    class Meta:
        namespace = "com.localhost"
        name = "SignalsTestKey"


@pytest.fixture
def ts() -> int:
    return 1632128298534


@pytest.fixture
def value_answer() -> bytes:
    return b'\x01\x00\x00\x00\x00\x00\x00\x06\x9c\x00\x00\x00\x01\xcc\xb8\xeb\xa6\x80_\x08test\x0cstring\x0cstring\x04NA'


@pytest.fixture
def test_producer() -> TestProducer:
    return TestProducer()


@pytest.fixture
def clean_producer(test_producer: TestProducer, sr_root: Path) -> HighLevelSerializingProducer:
    return HighLevelSerializingProducer(
        producer=test_producer,
        schema_registry=ClouderaSRClient(TestHTTPClient(sr_root), SimpleCache()),
        header_packer=ConfluentClouderaHeadersHandler().pack,
        serializer=AvroModelSerializer(),
        store=AvroModelRepo(),
        mapping={},
    )


def test_avro_producer_moving_parts_value_only(
    clean_producer: HighLevelSerializingProducer,
    test_producer: TestProducer,
    topic: str,
    ts: int,
    value_answer: bytes
) -> None:
    clean_producer.set_target_topic(topic, Signal)
    value = Signal(ts=ts)
    clean_producer.send_message(topic, value)
    [message] = test_producer.sent
    assert message.key is None
    assert message.value == value_answer


def test_avro_producer_moving_parts_value_and_key(
    clean_producer: HighLevelSerializingProducer,
    test_producer: TestProducer,
    value_answer: bytes,
    topic: str,
    ts: int,
) -> None:
    clean_producer.set_target_topic(topic, Signal, DefaultKey)
    key = DefaultKey(ts=ts)
    value = Signal(ts=ts)
    clean_producer.send_message(topic, value, key)
    [message] = test_producer.sent
    assert message.key == b'\x01\x00\x00\x00\x00\x00\x00\x06\x9d\x00\x00\x00\x01\xcc\xb8\xeb\xa6\x80_'
    assert message.value == value_answer



from typing import Dict, Any

import pytest
from pytest import FixtureRequest

from wunderkafka import SecurityProtocol

RawConfig = Dict[str, Any]


@pytest.fixture
def boostrap_servers() -> str:
    return 'localhost:9093'


@pytest.fixture
def default_config(boostrap_servers: str) -> RawConfig:
    return {'bootstrap_servers': boostrap_servers}


@pytest.fixture
def non_krb_config(boostrap_servers: str) -> RawConfig:
    return {
        'bootstrap_servers': boostrap_servers,
        'sasl_username': 'user',
        'sasl_password': 'password',
        'security_protocol': SecurityProtocol.sasl_ssl,
        'sasl_mechanism': 'SCRAM-SHA-512',
    }


@pytest.fixture(params=[SecurityProtocol.sasl_ssl, SecurityProtocol.sasl_plaintext])
def krb_config(request: FixtureRequest, boostrap_servers: str) -> RawConfig:
    return {
        'bootstrap_servers': boostrap_servers,
        'sasl_kerberos_keytab': 'user.keytab',
        'security_protocol': request.param,
    }
